//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/MakeScale.cpp
//! @brief     Implements functions that create instances of class Scale.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/MakeScale.h"
#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"

Scale GenericScale(const std::string& name, const std::vector<double>& limits)
{
    std::vector<Bin1D> vec;
    ASSERT(!(limits.size() & 1)); // must be an even number
    for (size_t i = 0; i < limits.size(); i += 2)
        vec.emplace_back(Bin1D::FromTo(limits[i], limits[i + 1]));
    return Scale(name, vec);
}

Scale* newGenericScale(const std::string& name, const std::vector<double>& limits)
{
    return new Scale(GenericScale(name, limits));
}


Scale ListScan(const std::string& name, const std::vector<double>& points)
{
    std::vector<Bin1D> vec;
    for (double p  : points)
        vec.emplace_back(Bin1D::At(p));
    return Scale(name, vec);
}

Scale* newListScan(const std::string& name, const std::vector<double>& points)
{
    return new Scale(ListScan(name, points));
}


Scale EquiDivision(const std::string& name, size_t N, double start, double end)
{
    ASSERT(N > 0);
    ASSERT(start <= end);
    std::vector<Bin1D> vec;
    for (size_t i = 0; i < N; ++i)
        vec.emplace_back(Bin1D::FromTo((N - i) * (start / N) + i * (end / N),
                                       (N - i - 1) * (start / N) + (i + 1) * (end / N)));
    ASSERT(vec.size() == N);

    return Scale(name, vec);
}

Scale* newEquiDivision(const std::string& name, size_t N, double start, double end)
{
    return new Scale(EquiDivision(name, N, start, end));
}

std::shared_ptr<Scale> sharedEquiDivision(const std::string& name, size_t N, double start,
                                          double end)
{
    return std::shared_ptr<Scale>(newEquiDivision(name, N, start, end));
}

std::unique_ptr<Scale> uniqueEquiDivision(const std::string& name, size_t N, double start,
                                          double end)
{
    return std::unique_ptr<Scale>(newEquiDivision(name, N, start, end));
}


Scale EquiScan(const std::string& name, size_t N, double start, double end)
{
    ASSERT(N);
    std::vector<Bin1D> vec;
    if (N == 1) {
        ASSERT(end == start);
        vec.emplace_back(Bin1D::At(start));
    } else {
        ASSERT(start <= end);
        for (size_t i = 0; i < N; ++i)
            vec.emplace_back(Bin1D::At((N - 1 - i) * (start / (N - 1)) + i * (end / (N - 1))));
        ASSERT(vec.size() == N);
    }

    return Scale(name, vec);
}

Scale* newEquiScan(const std::string& name, size_t N, double start, double end)
{
    return new Scale(EquiScan(name, N, start, end));
}

std::shared_ptr<Scale> sharedEquiScan(const std::string& name, size_t N, double start, double end)
{
    return std::shared_ptr<Scale>(newEquiDivision(name, N, start, end));
}

std::unique_ptr<Scale> uniqueEquiScan(const std::string& name, size_t N, double start, double end)
{
    return std::unique_ptr<Scale>(newEquiDivision(name, N, start, end));
}
